/***
 *
 * Copyright (C) 2003-2004 Fredrik Lingvall (fl@signal.uu.se)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ***/

/***
 *
 * Fredrik Lingvall 2004-01-21 : First release.
 *
 *
 ***/

#include "config.h"

#include "i18n.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include "filter.h"


// Set to 1 to force compiler to not use FFTW lib.
#if 0
#ifdef HAVE_LIBFFTW
#undef HAVE_LIBFFTW
#endif
#endif

#ifdef HAVE_LIBFFTW
#include <complex.h>
#include <fftw3.h>
#include <math.h>
#endif

static void init(void);
static void cleanup(void);
static int  mod_samples(gpointer *d, gint length, AFormat afmt, gint srate, gint nch);

#define MAX_FIR_LEN 8*4096	// Max FIR filter length.

EffectPlugin filter_ep =
  {
    NULL,
    NULL,
    "FIR Filter Plugin", /* Description */
    init,
    cleanup,
    filter_about,
    filter_configure,
    mod_samples,
    NULL
  };

static gint16  *buffer_l = NULL; // Circular buffer left channel vector.
static gint16  *buffer_r = NULL; // Circular buffer left channel vector.
static int     k;		// Circular buffer position.
static int     fir_len = 0;	// FIR filter lenth.

gboolean      reload_filter = TRUE;
gint          filter_gain = 10;	// Unity gain due to 10 x multiplication.
gchar         *filter_file  = NULL;
static double *h_l = NULL, *h_r = NULL; /* FIR filter tap vectors */

#ifdef HAVE_LIBFFTW

//
// FFTW stuff.
//

static int          fft_len = 0;
static int          blk_len = 0;

gboolean            fftw = TRUE; // Switch for using/not using FFTW for FIR filtering
static fftw_plan    p_forward_l   = NULL;
static fftw_plan    p_forward_r   = NULL;
static fftw_plan    p_backward_l  = NULL;
static fftw_plan    p_backward_r  = NULL;
static double      *u_l = NULL,  *u_r = NULL; // Input vectors.
static fftw_complex *uf_l = NULL,  *uf_r = NULL; // Fourier coefficents of input vectors.

static double      *olap_l = NULL,  *olap_r = NULL; // Overlap vectors.

static fftw_plan    p_forward_hl  = NULL;
static fftw_plan    p_forward_hr  = NULL;
static fftw_complex *hf_l = NULL,  *hf_r = NULL; // Fourier coefficents of the filters.

static gint         blk_pos;

#endif


EffectPlugin *get_eplugin_info(void)
{
  filter_ep.description = g_strdup_printf(_("FIR Filter Plugin %s"), VERSION);
  return &filter_ep;

}

static void init(void)
{
  ConfigFile *cfg;
  gint       n;

  // Allcate memory for filename.
  if (!filter_file) 
    filter_file = (char*) g_malloc0(256*sizeof(char));
  
  if (sizeof(short) != sizeof(gint16))
    abort();

  // Open the ~/.xmms/config file.
  cfg = xmms_cfg_open_default_file();

  // Read plugin paremeters.
  xmms_cfg_read_boolean(cfg, "filter_plugin", "reload_filter", &reload_filter);
  xmms_cfg_read_string(cfg, "filter_plugin", "filter_file", &filter_file);

  // Close the file.
  xmms_cfg_free(cfg);

  /* Allocate memory old samples */

  // Reset the circular buffer index.
  k = 0;

  if (!buffer_l)
    buffer_l = g_malloc0(sizeof(gint16)*(MAX_FIR_LEN));

  if (!buffer_r)
    buffer_r = g_malloc0(sizeof(gint16)*MAX_FIR_LEN);

  // Allocate space for FIR taps.

#ifndef HAVE_LIBFFTW

  printf("fftw3 not available! Using time-domain filtering.\n");
  
  if (!h_l) {
    h_l = (gdouble*) g_malloc0(sizeof(gdouble)*(MAX_FIR_LEN + 2)); 
    for (n=0; n<MAX_FIR_LEN; n++)
      h_l[n] = 0.0;
  }
  
  if (!h_r) {
    h_r = (gdouble*) g_malloc0(sizeof(gdouble)*(MAX_FIR_LEN + 2));
    for (n=0; n<MAX_FIR_LEN; n++)
      h_r[n] = 0.0;
  }
  
#else // FFTW stuff

#if 1 
  //printf("fftw_complex size = %d\n", sizeof(fftw_complex));
  //printf("double size = %d\n", sizeof(double));
  printf("fftw3 available! Initiating...");
#endif

  if (fftw) { // Frequency-domain method.
    
    // Allocate space for input vectors.
    
    if (!u_l) {
      u_l = (double*) fftw_malloc(sizeof(double)*MAX_FIR_LEN); 
      for (n=0; n<MAX_FIR_LEN; n++)
	u_l[n] = 0.0;
    }
    
    if (!u_r) {
      u_r = (double*) fftw_malloc(sizeof(double)*MAX_FIR_LEN);
      for (n=0; n<MAX_FIR_LEN; n++)
	u_r[n] = 0.0;
    }

    // Allocate space for FFT coefficients of input vectors (eg only fft_len/2 is needed).

    if (!uf_l) 
      uf_l = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*MAX_FIR_LEN); 
    
    if (!uf_r)
      uf_r = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*MAX_FIR_LEN);

    // Allocate space for overlap vectors.

    if (!olap_l) {
      olap_l = (double*) fftw_malloc(sizeof(double)*MAX_FIR_LEN); 
      for (n=0; n<MAX_FIR_LEN; n++)
	olap_l[n] = 0.0;
    }
    
    if (!olap_r) {
      olap_r = (double*) fftw_malloc(sizeof(double)*MAX_FIR_LEN);
      for (n=0; n<MAX_FIR_LEN; n++)
	olap_r[n] = 0.0;
    }

    // Allocate space for FIR taps.
  
    if (!h_l) {
      h_l = (double*) fftw_malloc(sizeof(double)*MAX_FIR_LEN); 
      for (n=0; n<MAX_FIR_LEN; n++)
	h_l[n] = 0.0;
    }
    
    if (!h_r) {
      h_r = (double*) fftw_malloc(sizeof(double)*MAX_FIR_LEN);
      for (n=0; n<MAX_FIR_LEN; n++)
	h_r[n] = 0.0;
    }

    // Allocate space for FFT coefficients of FIR taps (eg only fft_len/2 is needed).

    if (!hf_l) 
      hf_l = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*MAX_FIR_LEN); 
    
    if (!hf_r)
      hf_r = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*MAX_FIR_LEN);
    
  }
  else { // Time-domain method.
    
    // Allocate space for FIR taps.
    
    if (!h_l) {
      h_l = (double*) fftw_malloc(sizeof(double)*(MAX_FIR_LEN)); 
      for (n=0; n<MAX_FIR_LEN; n++)
	h_l[n] = 0.0;
    }
    
    if (!h_r) {
      h_r = (double*) fftw_malloc(sizeof(double)*(MAX_FIR_LEN));
      for (n=0; n<MAX_FIR_LEN; n++)
	h_r[n] = 0.0;
    }
  }

  printf("done.\n");

#endif

} // init

static void cleanup(void)
{
  
  printf("FIR Filter Plugin cleaning up...");

  if(filter_file) {
    g_free(filter_file);
    filter_file = NULL;
  }

#ifndef HAVE_LIBFFTW

  if (buffer_l) {
    g_free(buffer_l);
    buffer_l = NULL;	
  }
  
  if(buffer_r) {
    g_free(buffer_r);
    buffer_r = NULL;
  }	
  
  if(h_l) {
    g_free(h_l);
    h_l = NULL;
  }	
  
  if(h_r) {
    g_free(h_r);
    h_r = NULL;
  }	

#else

  // Free buffer memory.

  if (u_l) {
    fftw_free(u_l);
    u_l = NULL;
  }

  if (u_r) {
    fftw_free(u_r);
    u_r = NULL;
  }

  if (uf_l) {
    fftw_free(uf_l);
    uf_l = NULL;
  }
  
  if (uf_r) {
    fftw_free(uf_r);
    uf_r = NULL;
  }

  if (olap_l) {
    fftw_free(olap_l);
    olap_l = NULL;
  }
  
  if (olap_r) {
    fftw_free(olap_r);
    olap_r = NULL;
  }

  if (h_l) {
    fftw_free(h_l);
    h_l = NULL;
  }

  if (h_r) {
    fftw_free(h_r);
    h_r = NULL;
  }
  
  if (hf_l) {
    fftw_free(hf_l);
    hf_l = NULL;
  }

  if (hf_r) { 
    fftw_free(hf_r);
    hf_r = NULL;
  } 

  // Cleanup the plans.

  if (p_forward_l) {
    fftw_destroy_plan(p_forward_l);
    p_forward_l = NULL;
  } 
  
  if (p_forward_r) {
    fftw_destroy_plan(p_forward_r);
    p_forward_r = NULL;
  }
  
  if (p_backward_l) {
    fftw_destroy_plan(p_backward_l);
    p_backward_l = NULL;
  }

  if (p_backward_r) {
    fftw_destroy_plan(p_backward_r);
    p_backward_r = NULL;
  }

  if (p_forward_hl) {
    fftw_destroy_plan(p_forward_hl); 
    p_forward_hl = NULL;   
  }

  if (p_forward_hr) {
    fftw_destroy_plan(p_forward_hr); 
    p_forward_hr = NULL;
  }   

#endif

   printf("done!\n");
}

static int mod_samples(gpointer *d, gint length, AFormat afmt, gint srate, gint nch)
{
  register gint i, n;
  gint16        *data = (gint16 *) *d;
  static gint   old_srate, old_nch;
  FILE          *filterFile;
  char	        fileBuf[512];
  char          *endptr;
#ifndef HAVE_LIBFFTW
  register gdouble y_l, y_r, x;
  register gint m, in_l, in_r;
#endif
  //char	     *pBuf;

  if (!(afmt == FMT_S16_NE ||
	(afmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN) ||
	(afmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN)))
    return length;

  //
  // Flush the buffers if sampling rate has changed.
  //

  if (nch != old_nch || srate != old_srate)
    {
      memset(buffer_l, 0, sizeof(gint16)*MAX_FIR_LEN);
      memset(buffer_r, 0, sizeof(gint16)*MAX_FIR_LEN);
      k = 0;
      old_nch = nch;
      old_srate = srate;
    }

  //
  // Load filter coefficients from file.
  //

  if (reload_filter) {
    
    if (!(filterFile = fopen(filter_file, "r" ))) {
      fprintf(stderr, "FIR Plugin error:  Unable to open FIR coefficents file %s.\n",filter_file);
      return length;
    }
    
    n = 0;
    while ( fgets( fileBuf, sizeof( fileBuf ), filterFile )) {
      
      // Copy the filter parameters to memory. 
      h_l[n] = strtod(fileBuf, &endptr);
      h_r[n] = h_l[n];
      
      // Exit if max length exceded.    
      if (n >= MAX_FIR_LEN)
	break;
      
      n++;
    }
    
    if (n >= MAX_FIR_LEN)
      fir_len = MAX_FIR_LEN;
    else
      fir_len = n;

    printf("Filter length = %d\n", (int) fir_len);

    // Fill the rest of the FIR taps with zeros.
    if (n<MAX_FIR_LEN) { 
      for (; n<MAX_FIR_LEN; n++) {
	h_l[n] = 0.0;
	h_r[n] = 0.0;      
      }
    }
    
    if (filterFile)
      fclose(filterFile);

#ifdef HAVE_LIBFFTW

    if (fftw) { // Setup a new FFTW session.


      // Compute FFT and data block lengths.

      fft_len = MAX_FIR_LEN;
      blk_len = 128; // Min block length.
      n = 0;
      while ((int) pow(2.0,(double) n) < MAX_FIR_LEN) {
	if  ((int) pow(2.0, (double) n) >= fir_len + blk_len + 1) {
	  fft_len = (int) pow(2, (double) n);
	  break;
	}
	n++;
      }

      blk_len = fft_len + 1 - fir_len;
      //blk_len = fft_len - fir_len;

      // Check that blk_len is smaller than fragment size (=length/4).
      if (length/4 < blk_len) { 
	blk_len = length/4;
      }

      printf(" fft_len = %d\n blk_len = %d\n fir_len = %d\n", fft_len,blk_len,fir_len);

      // DFT

      printf("Updating fftw plans...");

      if (p_forward_l)
	fftw_destroy_plan(p_forward_l);
      p_forward_l = fftw_plan_dft_r2c_1d(fft_len, u_l, uf_l, FFTW_MEASURE);    

      if (p_forward_r)
	fftw_destroy_plan(p_forward_r);
      p_forward_r = fftw_plan_dft_r2c_1d(fft_len, u_r, uf_r, FFTW_MEASURE);    

      // IDFT

      if (p_backward_l)
	fftw_destroy_plan(p_backward_l);
      p_backward_l = fftw_plan_dft_c2r_1d(fft_len, uf_l, u_l, FFTW_MEASURE);    

      if (p_backward_r)
	fftw_destroy_plan(p_backward_r);
      p_backward_r = fftw_plan_dft_c2r_1d(fft_len, uf_r, u_r, FFTW_MEASURE);    

      // Fourier transform the filters.  

      if (p_forward_hl)
	fftw_destroy_plan(p_forward_hl);    
      p_forward_hl = fftw_plan_dft_r2c_1d(fft_len, h_l, hf_l, FFTW_MEASURE);
      fftw_execute(p_forward_hl);

      if (p_forward_hr)
	fftw_destroy_plan(p_forward_hr);    
      p_forward_hr = fftw_plan_dft_r2c_1d(fft_len, h_r, hf_r, FFTW_ESTIMATE);
      fftw_execute(p_forward_hr);

      printf("done.\n");

#if 0
      // Print the frequency response of the filter. 
      for (n = 0; n < fft_len; n++) {
	printf("%g %g\n",crealf( hf_l[n]), cimagf( hf_l[n]));
      }
#endif
      //fftw_export_wisdom_to_file(stdout);

    }
#endif // HAVE_LIBFFTW  
    
    reload_filter = FALSE;
  }


#ifndef HAVE_LIBFFTW 

  //
  // Precess next sample (time domain).
  //
  
  //printf("Length = %d\n",(int) length);

  for (i = 0; i < length/2; i+=2) {
    
    in_l = data[i];		// Get a new left channel sample.
    in_r = data[i+1];		// Get a new right channel sample.

    // Start from zero if the buffer is full.
    if (k >= fir_len)
      k = 0;
    
    // Store the new samples in the circular buffer.
    buffer_l[k] = in_l;
    buffer_r[k] = in_r;

    y_l = 0.0; 
    y_r = 0.0; 
    for (n=0; n<fir_len; n++) { // FIR Update.
    
      //
      // Compute circular buffer index (the index m=k-n in the FIR update y[k] += h[n]*x[k-n]).
      //  
       
      if ( (k-n) < 0)
	m = (k-n) + fir_len;
      else
	m = k-n;
      
      //
      // Left channel.
      //
      
      x = (gdouble) buffer_l[m];
      y_l += h_l[n] * x;
      
      //
      // Right channel.
      //
      
      x = (gdouble) buffer_r[m];
      y_r += h_r[n] * x;

     }
    
    y_l *= (gdouble) filter_gain / 10.0;
    y_r *= (gdouble) filter_gain / 10.0;

    // Limit the output to 16 bits.
    y_l = CLAMP(y_l, -32768, 32767);
    y_r = CLAMP(y_r, -32768, 32767);
    
    // Output the filtered data.
    data[i]   = (gint16) y_l;
    data[i+1] = (gint16) y_r;
    
    // Update buffer index.
    k++;
  }

#else // HAVE_LIBFFTW (libfftw present). 

  //
  // Frequency domain method (overlap-and-add algorithm).
  //


  // Check if in-buffer is shorter than blk_len.
  if (length/4 < blk_len) { 
    blk_len = length/4;
    //printf("blk_len reduced to %d\n",blk_len);
  }
  
  // Process a new data block.
  for (blk_pos = 0; blk_pos<length/2; blk_pos += 2*blk_len) { // 2*blk_len since in-buffer is in bytes.
    
    //    
    // Read in a new data block from input stream.
    //
    
    for (i=0, n=0; n < fft_len; i+=2, n++) {
      if (blk_pos+i < length/2  && n < blk_len ) {
	u_l[n] = (double) data[blk_pos + i]; // Get a new left channel sample.
	u_r[n] = (double) data[blk_pos + i+1]; // Get a new right channel sample.
      }
      else { // Zero-pad
	u_l[n] = 0.0;
	u_r[n] = 0.0;
      } // if
    } // for

#if 0
    printf("******** BEFORE ******\n");
    for (k=0; k<fft_len; k++)
      printf("%g\n",  (double) u_l[k]);
#endif

    //
    // Compute DFT of the new data.
    //
    
    // Left channel.
    //fftw_execute_dft_r2c(p_forward_l, u_l, uf_l);
    fftw_execute(p_forward_l);
    
    // Right channel.
    //fftw_execute_dft_r2c(p_forward_r, u_r, uf_r);
    fftw_execute(p_forward_r);

    // Do the filtering.
    for (n = 0; n < fft_len/2+1; n++) {
      uf_l[n] *= hf_l[n] / (double) fft_len;
      uf_r[n] *= hf_r[n] / (double) fft_len;
    }

    //
    // Compute the inverse DFT of the filtered data.
    //

    // Left channel.
    //fftw_execute_dft_c2r(p_backward_l, uf_l, u_l);
    fftw_execute(p_backward_l);

    // Right channel.
    //fftw_execute_dft_c2r(p_backward_r, uf_r, u_r);
    fftw_execute(p_backward_r);

#if 0
    printf("******** AFTER ******\n");
    for (k=0; k<fft_len; k++)
      printf("%g\n",  (double) u_l[k]);
#endif

#if 0
    for (k=0; k<fft_len; k++)
      if (fabs(u_l[k]) > 32768.0)
	printf("Clipping ifft! (%d)\n",k);
#endif

    //
    // Overlap-and-Add
    //
      
    if ( (length/2-blk_pos)/2 >= blk_len) // Have blk_len samples in buffer.
      k = blk_len;
    else		  // Have less than blk_len samples in buffer.
      k = (length/2-blk_pos)/2;
    
    // Add the last block's overlap to the current one.
    for (n = 0; n < fir_len-1; n++) {
      u_l[n] += olap_l[n]; 
      u_r[n] += olap_r[n];
    }

    // Save the samples that will overlap to the next block.
    for (n = k; n < k+fir_len; n++) {
      olap_l[n - k] = u_l[n]; 
      olap_r[n - k] = u_r[n]; 
    }

    // Copy the processed data.
    for (i = 0,n=0; n < blk_len; i+=2,n++) {

      if (blk_pos+i < length/2) {
	
	// Left channel.
	data[blk_pos + i]   =  CLAMP( (gint16) u_l[n], -32768, 32767);
	
	// Right channel.
	data[blk_pos + i+1] =  CLAMP( (gint16) u_r[n], -32768, 32767);

      } // if
    } // for
    
    
  } // for (blk_pos ...

  // Restore blk_len
  blk_len = fft_len + 1 - fir_len;

#endif // HAVE_LIBFFTW  

  return length;
}

